package com.huguesjohnson;

/*
DataFile.java - Reads data from a file into a neural network layer
Copyright  (C) 2000-2003 Hugues Johnson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
//import java.lang.Double;
//import java.lang.NumberFormatException;
import java.lang.Object;
import java.lang.String;
import java.util.StringTokenizer;

/** class DataFile
 */
public class DataFile extends Object{
    /** reads the file */    
    private LineNumberReader reader;
    
    /** the current line in memory */    
    private String currentLine;
    
    /** the next line in memory */    
    private String nextLine;
    
    /** comment character */    
    private String comment;
    
    /** field delimiter */    
    private String delimiter;
    
    /** default comment character */
    private final static String DEFAULT_COMMENT="//";

    /** default delimiter character */
    private final static String DEFAULT_DELIMITER=",";
    
    /** constructor, uses default comment and delimiter characters
     * @param fileName name & path of data file to open
     */
    public DataFile(String fileName){
        this.comment=this.DEFAULT_COMMENT;
        this.delimiter=this.DEFAULT_DELIMITER;
        this.openReader(fileName);
    }
    
    /** constructor, uses default comment and delimiter characters
     * @param fileName name & path of data file to open
     * @param comment the comment character or string for the file
     * @param delimiter the delimiter character or string for the file
     */
    public DataFile(String fileName,String comment,String delimiter){
        this.comment=comment;
        this.delimiter=delimiter;
        this.openReader(fileName);
    }    
    
    /** opens the reader
     * @param fileName name & path of data file to open
     */
    private void openReader(String fileName){
        try{
            this.reader=new LineNumberReader(new FileReader(fileName));
        } catch(FileNotFoundException fnfx){
            fnfx.printStackTrace();
        }
        if(this.reader!=null){
            try{
                this.currentLine=this.reader.readLine();
                 while(this.isCommentLine(this.currentLine)){
                    this.currentLine=this.reader.readLine();
                }
                this.nextLine=this.reader.readLine();
                while(this.isCommentLine(this.nextLine)){
                    this.nextLine=this.reader.readLine();
                }
            } catch(IOException iox){
                iox.printStackTrace();
            }
        }
    }
    
    /** isCommentLine
     * @return true if line is a comment
     * @param line the line to test
     */
    private boolean isCommentLine(String line){
        return(line.startsWith(this.comment));
    }

    /** reads ahead a line
     * @return the line that was read
     */    
    private StringTokenizer readNext(){
        StringTokenizer tokenizer=new StringTokenizer(this.currentLine,this.delimiter);
        this.currentLine=this.nextLine;
        if(this.reader!=null){
            try{
                this.nextLine=this.reader.readLine();
                while(this.isCommentLine(this.nextLine)){
                    this.nextLine=this.reader.readLine();
                }
            } catch(IOException iox){
                iox.printStackTrace();
            } finally{
                return(tokenizer);
            }
        }
        return(tokenizer);
    }
    
    /** endOfFile
     * @return true if at the end of the file
     */
    public boolean endOfFile(){
        return(this.currentLine==null);
    }
    
    /** getNextLine, retrieves the next line of data
     * @return double array containing a line of data
     */
//    public double[] getNextLine(){
//        double[] dArray=null;
//        int tokenCount=0;
//        StringTokenizer tokenizer=this.readNext();
//        if(tokenizer!=null){
//            tokenCount=tokenizer.countTokens();
//        }
//        if(tokenCount>0){
//            dArray=new double[tokenCount];
//            for(int index=0;index<tokenCount;index++){
//                try{
//                    dArray[index]=Double.valueOf(tokenizer.nextToken()).doubleValue();
//                } catch(NumberFormatException nfx){
//                    dArray[index]=0.0D;
//                }
//            }
//        }
//        return(dArray);
//    }

    /** getNextLine, retrieves the next line of data
     * @return String array containing a line of data
     */
    public String[] getNextLine(){
        String[] sArray=null;
        int tokenCount=0;
        StringTokenizer tokenizer=this.readNext();
        if(tokenizer!=null){
            tokenCount=tokenizer.countTokens();
        }
        if(tokenCount>0){
            sArray=new String[tokenCount];
            for(int index=0;index<tokenCount;index++){
                sArray[index]=tokenizer.nextToken();
            }
        }
        return(sArray);
    }    
    
    /** closes the file
     */
    public void close(){
        try{
            this.reader.close();
        } catch(IOException iox){
            iox.printStackTrace();
        }
    }
    
    /** toString
     * @return string representation of object
     */
    public String toString(){
        StringBuffer tostring=new StringBuffer();
        tostring.append(this.getClass().getName());
        tostring.append("\nreader="+this.reader.toString());
        tostring.append("\ncurrentLine="+this.currentLine);
        tostring.append("\nnextLine="+this.nextLine);
        tostring.append("\ncomment="+this.comment);
        tostring.append("\ndelimiter="+this.delimiter);
        return(tostring.toString());
    }    
}