package com.huguesjohnson.jail.jdt;

/*
Jdt (Java Decision Tree) - Package to create and display a decision tree
DecisionTreeNode.java - node for a decision tree
Copyright  (C) 2000-2003 Hugues Johnson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

import java.io.Serializable;
import java.lang.Cloneable;
import java.lang.CloneNotSupportedException;
import java.lang.Object;
import java.lang.String;

/** JdtDecisionTreeNode, representation for tree node
*/
public class DecisionTreeNode implements Cloneable,Serializable{
    /* constants */
    /** left child index */
    public static final int CHILD_LEFT=0;
    /** right child index */
    public static final int CHILD_RIGHT=1;
    /** not used in current project, but could be used for a more advanced project */
    public static final int OP_EQUAL=0; 
    /** only operation used at present time */
    public static final int OP_GREATER_THAN=1;
    /** not used in current project, but could be used for a more advanced project */
    public static final int OP_LESS_THAN=2; 
    /* private shared members */
    /** parent node */
    private DecisionTreeNode parent;
    /** left child node */
    private DecisionTreeNode leftChild;
    /** right child node */
    private DecisionTreeNode rightChild;
    /** classification to return if a leaf node is reached */
    private String classification;
    /** attribute that is tested at the node */
    private int attributeIndex; 
    /** value that is compared at the node */
    private double compareValue; 
    /** operator used in comparision at node */
    private int operator; 
    
    /** constructor, creates a node with no parent
     */
    public DecisionTreeNode(){
        this.parent=null;
        this.leftChild=null;
        this.rightChild=null;
    }
    
    /** constructor, creates node with a parent
     * @param parent parent node
     */
    public DecisionTreeNode(DecisionTreeNode parent){
        this.parent=parent;
        this.leftChild=null;
        this.rightChild=null;
    }
    
    /** clone
     * @return clone of object
     */
    public Object clone(){
        try{
            return(super.clone());
        } catch(CloneNotSupportedException cnsx){
            return(null);
        }
    }
    
    /** determines if node is a root node
     * @return true if node is a root node (no parent)
     */
    public boolean isRoot(){
        return(this.parent==null);
    }
    
    /** determines if node is a leaf node
     * @return true if node is a leaf node (no no children)
     */
    public boolean isLeaf(){
        return((this.leftChild==null)&&(this.rightChild==null));
    }
    
    /** appends a child to the current node
     * @param direction CHILD_LEFT or CHILD_RIGHT
     * @param node the node to attach
     */
    public void setChildNode(int direction,DecisionTreeNode node){
        if(direction==CHILD_LEFT){
            this.leftChild=node;
        } else if (direction==CHILD_RIGHT){
            this.rightChild=node;
        }
    }
    
    /** retrieves the parent node
     * @return parent node
     */
    public DecisionTreeNode getParentNode(){
        return(this.parent);
    }
    
    /** retrieves a child node
     * @param direction which child to retrieve
     * @return tree node or null if no child exists
     */
    public DecisionTreeNode getChildNode(int direction){
        DecisionTreeNode returnNode=null;
        if(direction==CHILD_LEFT){
            return(this.leftChild);
        } else if(direction==CHILD_RIGHT){
            return(this.rightChild);
        }
        return(returnNode);
    }
    
    /** sets classification for node
     * @param classification new classification for node
     */
    public void setClassification(String classification){
        this.classification=classification;
    }
    
    /** retreives classification of node
     * @return classification of node
     */    
    public String getClassification(){
        return(this.classification);
    }
    
    /** sets attribute index for node
     * @param attributeIndex new attribute index for node
     */
    public void setAttributeIndex(int attributeIndex){
        this.attributeIndex=attributeIndex;
    }
    
    /** retrieves attribute index for current node
     * @return attribute index
     */
    public int getAttributeIndex(){
        return(this.attributeIndex);
    }    
    
    /** sets compare value for node
     * @param compareValue new compare value for node
     */
    public void setCompareValue(double compareValue){
        this.compareValue=compareValue;
    }
    
    /** retrieves compare value for current node
     * @return compare value
     */
    public double getCompareValue(){
        return(this.compareValue);
    }

    /** sets comparision operator for node
     * @param operator new comparision operator for node
     */
    public void setCompareOperator(int operator){
        this.operator=operator;
    }
    
    /** retrieves comparision operator for current node
     * @return compare operator
     */
    public int getCompareOperator(){
        return(this.operator);
    }
    
    /** returns a string representation of the node 
     * @return representation of node
     */
    public String toString(){
        if(this.isLeaf()){
            return(this.classification);
        } else if(this.operator==OP_GREATER_THAN){
            return("a"+this.attributeIndex+">"+this.compareValue);
        } else if(this.operator==OP_LESS_THAN){
            return("a"+this.attributeIndex+"<"+this.compareValue);
        } else {
            return("a"+this.attributeIndex+"="+this.compareValue);
        }
    }
}