package com.huguesjohnson.jail.jdt;

/*
Jdt (Java Decision Tree) - Package to create and display a decision tree
Jdt.java - entry point for the program 
Copyright  (C) 2000-2003 Hugues Johnson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

import java.awt.event.ActionEvent;
import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.lang.Exception;
import java.lang.Object;
import java.lang.String;
import java.lang.System;
import javax.swing.JOptionPane;
import com.huguesjohnson.IniFile;
import com.huguesjohnson.jail.jdt.JdtApp;
import com.huguesjohnson.jail.jdt.JdtUIConstants;
import com.huguesjohnson.jail.jdt.JdtUIWindow;
import com.huguesjohnson.ui.SplashScreen;

/** class Jdt
 * @author Hugues Johnson
 */
public class Jdt extends Object{
    /** message that displays if load fails */
    private final static String ERROR_UNABLE_TO_LOAD="Unable to load program. \nPlease check that the latest version of the Java Runtime Environment is installed.";
    
    /** name of the program */
    private final static String PROGRAM_NAME="Jdt";
    
    /** version of the program */
    private final static String PROGRAM_VERSION="0.1";
    
    /** information displayed in the about dialog */
    private final static String PROGRAM_INFO=PROGRAM_NAME+" "+PROGRAM_VERSION+"\nCopyright  (C) 2000-2002\nHugues Johnson";
    
    /** reference to the main window */
    private JdtUIWindow mainWindow;
    
    /** reference to the underlying decision tree application */
    private JdtApp jdtapp;
    
    /** flag to log errors */
    private boolean logErrors=false;
    
    /** flag to exit Java runtime systems on application close */
    private boolean exitSystemOnClose;
    
    /** constructor
     * @param args command line arguments
     * @param exitSystemOnClose flag to exit Java runtime system on application close
     */
    public Jdt(String args[],boolean exitSystemOnClose){
        this.exitSystemOnClose=exitSystemOnClose;
        if(this.logErrors){
            this.setErrorStream();
        }
        this.createWindow();
        /* make sure window has been created successfully */
        if(this.mainWindow==null){
            System.err.println(ERROR_UNABLE_TO_LOAD);
            System.exit(0);
        }
        SplashScreen splash=new SplashScreen(this.mainWindow,this.mainWindow.getResourcePath()+"splash.gif");
        this.initWindow();
        splash.dispose();
    }
    
    /** creates the main window
     */
    private void createWindow(){
        mainWindow=new JdtUIWindow("Jdt"){
            public void actionListener(final ActionEvent e){
                actionHandler(e);
            }
            public void closeWindow(){
                unloadWindow();
            }
        };
    }
    
     /** invoked on the windowClosing event from the main window
      */
    private final void unloadWindow(){
        this.saveSettings();
        this.mainWindow.dispose();
        if(this.exitSystemOnClose){
            System.exit(0);
        }
    }
    
    /** intializes mainWindow and shows it
     */
    private final void initWindow(){
        this.mainWindow.pack();
        this.mainWindow.setSize(600,400);
        this.mainWindow.setVisible(true);
        this.loadSettings();
    }
    
    /** receives all actions and parses for source and actionCommand.
     * A message is then created and sent to the message handler.
     * @param e ActionEvent sent to the UI
     * @TODO split command parser into another routine to allow command line support - just use this to intercept messages
     */
    private final void actionHandler(ActionEvent e){
        String actionCommand=e.getActionCommand();
        if(actionCommand.equals(JdtUIConstants.ACTION_SET_PANE_WELCOME)){
            this.mainWindow.setRightPane(this.mainWindow.PANE_WELCOME);
        } else if(actionCommand.equals(JdtUIConstants.ACTION_SET_PANE_FILE)){
            this.mainWindow.setRightPane(this.mainWindow.PANE_FILE);
        } else if(actionCommand.equals(JdtUIConstants.ACTION_SET_PANE_RUN)){
            this.mainWindow.setRightPane(this.mainWindow.PANE_RUN);
        } else if(actionCommand.equals(JdtUIConstants.ACTION_SET_PANE_TREE)){
            this.mainWindow.setRightPane(this.mainWindow.PANE_TREE);
        } else if(actionCommand.equals(JdtUIConstants.ACTION_RUN)){
            this.runDecisionTreeLearn();
        } else if(actionCommand.equals(JdtUIConstants.ACTION_EXIT)){
            this.unloadWindow();
        } else if(actionCommand.equals(JdtUIConstants.ACTION_ABOUT)){
            this.showAboutDialog();
        }
    }
    
    /** receives status messages from jdtapp
     * @param message message sent from jdtapp
     */
    private void updateProgress(String message){
        this.mainWindow.setProgress(message,0);
    }

    /** receives status messages from jdtapp
     * @param message message sent from jdtapp
     * @param percent precent of current task complete
     */
    private void updateProgress(String message,int percent){
        this.mainWindow.setProgress(message,percent);
    }
    
    /** called when jdtapp thread is finished
     */
    private void treeDone(){
        this.mainWindow.setProgress(JdtUIConstants.STATUS_DONEBUILD,0);
    }
    
    /** shows an error dialog
     * @param x error received
     */
    private void errorMessage(Exception x){
        /* show error message */
        JOptionPane.showMessageDialog(this.mainWindow,x.toString(),"Error",JOptionPane.ERROR_MESSAGE);
        /* log error */
        if(this.logErrors){
            System.err.println(System.currentTimeMillis());
            x.printStackTrace();
        }
    }

    /** shows an about dialog
     */
    private void showAboutDialog(){
        /* show dialog */
        JOptionPane.showMessageDialog(this.mainWindow,this.PROGRAM_INFO,this.PROGRAM_NAME,JOptionPane.INFORMATION_MESSAGE);
    }    
    
     /** runDecisionTreeLearn
      */
    private void runDecisionTreeLearn(){
        this.mainWindow.setProgress(JdtUIConstants.STATUS_BUILDTREE,0);
        boolean start=true;
        String namesFile=this.mainWindow.getNamesFileName();
        String trainingFile=this.mainWindow.getTrainingFileName();
        String testFile=this.mainWindow.getTestFileName();
        /* names, training, & testing files must exist to continue - log & report files are optional */
        if((new File(namesFile)).exists()==false){
            this.errorMessage(new Exception(JdtUIConstants.ERROR_FILE_NOT_FOUND+namesFile));
            start=false;
        } else if((new File(trainingFile)).exists()==false){
            this.errorMessage(new Exception(JdtUIConstants.ERROR_FILE_NOT_FOUND+trainingFile));
            start=false;
        } else if((new File(testFile)).exists()==false){
            this.errorMessage(new Exception(JdtUIConstants.ERROR_FILE_NOT_FOUND+testFile));
            start=false;
        }
        if(start==true){
            this.jdtapp=new JdtApp(namesFile,trainingFile,testFile,this.mainWindow.getDebugFileName(),this.mainWindow.getReportFileName()){
                public void updateStatus(String message){
                    updateProgress(message);
                }
                public void done(){
                    treeDone();
                }
                public void error(Exception x){
                    errorMessage(x);
                }
            };
            this.jdtapp.start();
        } else{
            this.mainWindow.setProgress(JdtUIConstants.STATUS_SELECT_FILES,0);
        }
    }
    
    /** creates a text file to log run-time errors
     */
    private void setErrorStream(){
        try{
            System.setErr(new PrintStream(new FileOutputStream(System.getProperty("user.dir")+File.separator+this.PROGRAM_NAME+".err",true)));
        } catch(Exception x){
            x.printStackTrace();
        }
    }
    
    /** saves program settings to an IniFile
     */
    private void saveSettings(){
        IniFile iniFile=new IniFile(System.getProperty("user.dir")+File.separator+"settings"+File.separator+"jdt.ini");
        if(iniFile!=null){
            iniFile.setProperty("namesFile",this.mainWindow.getNamesFileName());
            iniFile.setProperty("trainFile",this.mainWindow.getTrainingFileName());
            iniFile.setProperty("testFile",this.mainWindow.getTestFileName());
            iniFile.setProperty("reportFile",this.mainWindow.getReportFileName());
            iniFile.setProperty("debugFile",this.mainWindow.getDebugFileName());
            iniFile.save(this.PROGRAM_NAME+" "+this.PROGRAM_VERSION);
            iniFile=null;
        }    
    }
    
    /** loads settings from an IniFile
     */
    private void loadSettings(){
        IniFile iniFile=new IniFile(System.getProperty("user.dir")+File.separator+"settings"+File.separator+"jdt.ini");
        if(iniFile!=null){
            iniFile.read();
            String property=new String();
            final String defaultString=new String("__dEfAulT__");
            property=iniFile.getProperty("namesFile",defaultString);
            if(property.equals(defaultString)==false){
                this.mainWindow.setNamesFileName(property);
            }
            property=iniFile.getProperty("trainFile",defaultString);
            if(property.equals(defaultString)==false){
                this.mainWindow.setTrainingFileName(property);
            }
            property=iniFile.getProperty("testFile",defaultString);
            if(property.equals(defaultString)==false){
                this.mainWindow.setTestFileName(property);
            }
            property=iniFile.getProperty("reportFile",defaultString);
            if(property.equals(defaultString)==false){
                this.mainWindow.setReportFileName(property);
            }            
            property=iniFile.getProperty("debugFile",defaultString);
            if(property.equals(defaultString)==false){
                this.mainWindow.setDebugFileName(property);
            }            
            iniFile.clear();
            iniFile=null;
          }
    }
    
    /** returns a string representation of the object
     * @return string representation of the object
     */
    public String toString(){
        String tostring=new String(super.toString());
        return(tostring);
    }    
    
    /** entry point for program
     * @param args command line arguments
     */
    public static void main(String args[]){
        try{
            Jdt instance=new Jdt(args,true);
        } catch(Exception x){
            System.err.println(ERROR_UNABLE_TO_LOAD);
            x.printStackTrace();
        }
    }
}