package com.huguesjohnson.jail.jdt;

/*
Jdt (Java Decision Tree) - Package to create and display a decision tree
JdtApp.java - main program, handles interaction between window and decision tree
Copyright  (C) 2000-2003 Hugues Johnson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.lang.ClassNotFoundException;
import java.lang.Double;
import java.lang.Integer;
import java.lang.String;
import java.lang.System;
import java.lang.Thread;
import java.util.StringTokenizer;
import java.util.Vector;
import com.huguesjohnson.jail.jdt.Attribute;
import com.huguesjohnson.jail.jdt.DataInstance;
import com.huguesjohnson.jail.jdt.DecisionTree;
import com.huguesjohnson.DirectoryFilter;
import com.huguesjohnson.LogFile;

/**class JdtApp
 * @author Hugues Johnson
 */
public class JdtApp extends Thread{
    
    /** error message given if names file is in icorrect format */
    private final String errorBadNamesFile="Error, names file is in an incorrect format. The first line must contain classifications. The second line may be blank. Each remaining line must be an attribute name.";
    
    /** error message given if log file can not be created */
    private final String errorFailCreateLog="Error, can't create log file. Continuing execution..";
    
    /** error message given if a file is not found */
    private final String errorFileNotFound="Error, file not found: ";
    
    /** status message given when tree is being built */
    private final String statusBuildTree="Building initial decision tree, this may take a couple-o-mintues..";
    
    /** status message given when program is closing */
    private final String statusClosingProgram="Finishing decision tree build..";
    
    /** status message given when initializing decision tree build */
    private final String statusInitialize="Intializing decision tree build..";
    
    /** status message given when pruning tree */
    private final String statusPruning="Pruning initial decision tree, this may take a couple-o-minutes..";
    
    /** status message given when reading attributes */
    private final String statusReadingAttributes="Reading attributes and building attributes vector..";
    
    /** status message given when reading classifications */
    private final String statusReadingClassifications="Reading classifications and building classification list..";
    
    /** status message given when testing the accuracy of the trees */
    private final String statusTestingAccuracy="Testing accuracy of trees..";
    
    /** status message given when writing the report */
    private final String statusWritingReport="Writing report..";
    
    /** token for field delimiter in names file */
    private final String fieldDelimeter=",";

    /** path to the names file */
    private String namesFile;
    
    /** path to the training data file */
    private String trainingFile;
    
    /** path to the test data */
    private String testFile;
    
    /** path to the report file */
    private String reportFile;
    
    /** log file */
    private LogFile logFile;
    
    /** whether or not to create a log file */
    private boolean createLog;
    
    /** whether or not to create a report file */
    private boolean createReport;
    
    /** start time of execution */
    private long startTime;
    
    /** array of valid classifications */
    private String classifications[];
    
    /** vector of attributes */
    private Vector attributes;
    
    /** vector of training data */
    private Vector trainingData;
    
    /** initial capacity for attribute vector */
    private final int initAttributeCapacity=5;
    
    /** initial capacity increment for attribute vector */
    private final int initAttributeIncrement=10;
    
    /** initial decision tree */
    private DecisionTree initialTree;
    
    /** pruned decision tree */
    private DecisionTree prunedTree;
    
    /** percent correct of test data on initial tree */
    private double percentCorrectInitialTreeTest;
    
    /** percent correct of test data on pruned tree */
    private double percentCorrectPrunedTreeTest;
    
    /** percent correct of pruning data on initial tree */
    private double percentCorrectInitialTreePrune;
    
    /** percent correct of pruning data on pruned tree */
    private double percentCorrectPrunedTreePrune;
    
    /** percent correct of training data on initial tree */
    private double percentCorrectInitialTreeTrain;
    
    /** percent correct of training data on pruned tree */
    private double percentCorrectPrunedTreeTrain;
    
    /** constructor
     * @param namesFile path to names file
     * @param trainingFile path to training file
     * @param testFile path to test file
     * @param logFile path to log file
     * @param reportFile path to report file
     */
    public JdtApp(String namesFile,String trainingFile,String testFile,String logFile,String reportFile){
        this.namesFile=namesFile;
        this.trainingFile=trainingFile;
        this.testFile=testFile;
        if(logFile==null){
            this.createLog=false;
        } else{
            this.createLog=true;
            this.openLogFile(logFile);
        }
        if(reportFile==null){
            this.createReport=false;
        } else{
            this.createReport=true;
            this.reportFile=reportFile;
        }
    }
    
    /** starts decision tree learn
     */    
    public void run(){
        //this.createLog=false;
        this.startTime=System.currentTimeMillis();
        //if(this.createLog){
        //    this.openLogFile();
        //}
        this.readNamesFile(this.namesFile);
        this.readTrainFile(this.trainingFile);
        this.readTestFile(this.testFile);
        this.endProgram();
        this.done();
    }
    
    /** close any open files and terminate thread execution
     */
    private void endProgram(){
        /* print trees */
        //this.updateStatus("\nPrinting trees..");
        //this.updateStatus("\nBreadth First Search of Initial Tree:\n"+this.initialTree.breadthFirstSearch());
        //this.updateStatus("\nContents of Initial Tree:\n"+this.initialTree.toString());
        /* show statistics */
        //this.updateStatus("\nPercent of correct classifications");
        //this.updateStatus("Training Data on Initial Tree "+this.percentCorrectInitialTreeTrain+"%");
        //this.updateStatus("Training Data on PrunedTree "+this.percentCorrectPrunedTreeTrain+"%");
        //this.updateStatus("Pruning Data on Initial Tree "+this.percentCorrectInitialTreePrune+"%");
        //this.updateStatus("Pruning Data on Pruned Tree "+this.percentCorrectPrunedTreePrune+"%");
        //this.updateStatus("Test Data on Initial Tree "+this.percentCorrectInitialTreeTest+"%");
        //this.updateStatus("Test Data on Pruned Tree "+this.percentCorrectPrunedTreeTest+"%");
        this.log("Ended execution at: "+System.currentTimeMillis());
        if(this.createLog){
            if(this.logFile!=null){
                this.logFile.close();
                this.updateStatus("Created or appended execution log" /*: Project2.log"*/);
            }
        }
        if(this.createReport){
            this.updateStatus(this.statusWritingReport);
            this.generateReport();
            this.updateStatus("Created or appended report"/*: Project2.rpt"*/);
        }
        this.updateStatus(this.statusClosingProgram);
        this.attributes.clear(); 
    }
    
    /** create execution log
     * @param fileName file name
     */
    private void openLogFile(String fileName){
        this.logFile=new LogFile(fileName);
        if(this.logFile==null){
            this.updateStatus(this.errorFailCreateLog);
            this.createLog=false;
        }
    }
    
    /** create final statistical report
     */
    private void generateReport(){
        LogFile reportFile=new LogFile(this.reportFile);
        int index=0;
        if(reportFile!=null){
            /* log classifications */
            if(this.classifications!=null){
                reportFile.log("");
                reportFile.log("Classifications");
                String classificationsLine=new String("");
                for(index=0;index<classifications.length;index++){
                    classificationsLine=classificationsLine.concat(classifications[index]+" ");
                }
                reportFile.log(classificationsLine);
            }
            /* log contents of tree */
            reportFile.log("");
            reportFile.log("Breadth First Search of Initial Tree: "+this.initialTree.breadthFirstSearch());
            //reportFile.log("Breadth First Search of Pruned Tree: "+this.prunedTree.breadthFirstSearch());
            reportFile.log("");
            reportFile.log("Initial Tree");
            reportFile.log(this.initialTree.toString());
            reportFile.log("");
            //reportFile.log("Pruned Tree");
            //reportFile.log(this.prunedTree.toString());
            /* log accuracy */
            reportFile.log("");
            reportFile.log("Accuracy of decision trees");
            reportFile.log("Training Data on Initial Tree "+this.percentCorrectInitialTreeTrain+"%");
            //reportFile.log("Training Data on Pruned Tree "+this.percentCorrectPrunedTreeTrain+"%");
            //reportFile.log("Pruning Data on Initial Tree "+this.percentCorrectInitialTreePrune+"%");
            //reportFile.log("Pruning Data on Pruned Tree "+this.percentCorrectPrunedTreePrune+"%");
            reportFile.log("Test Data on Initial Tree "+this.percentCorrectInitialTreeTest+"%");
            //reportFile.log("Test Data on Pruned Tree "+this.percentCorrectPrunedTreeTest+"%");
            /* log run time */
            reportFile.log("");
            reportFile.log("Total program runtime: "+(((double)System.currentTimeMillis()-(double)this.startTime)/1000.0D)+" seconds");
            /* close file */
            reportFile.close();
        }
    }
    
    /** reads and parses names file
     * @param dataSetName name of data set to use
     */
    private void readNamesFile(String dataSetName){
        int index;
        this.updateStatus("Processing "+dataSetName+"..");
        LineNumberReader reader=null;
        StringTokenizer tokenizer;
        try{
            reader=new LineNumberReader(new FileReader(dataSetName));
        } catch(Exception x){
            /* log error and quit */
            this.error(x);
            this.log(x.toString());
            this.updateStatus(this.errorFileNotFound+dataSetName);
            this.endProgram();
        }
        if(reader!=null){
            try{
                this.updateStatus(this.statusReadingClassifications);
                /* first line of .names file contains legal classifications */
                String classificationsLine=reader.readLine().trim();
                /* chop off trailing "." at the end of classificationsLine */
                if(classificationsLine.endsWith(".")){
                    classificationsLine=classificationsLine.substring(0,classificationsLine.length()-1);
                }
                tokenizer=new StringTokenizer(classificationsLine,this.fieldDelimeter);
                int tokenCount=tokenizer.countTokens();
                this.classifications=new String[tokenCount];
                for(index=0;index<tokenCount;index++){
                    this.classifications[index]=tokenizer.nextToken().trim();
                    this.log("Classification"+index+"="+this.classifications[index]);
                }
                /* read attribute names, one attribute per line, mark current position because read() call jumps ahead one character */
                this.updateStatus(this.statusReadingAttributes);
                this.attributes=new Vector(this.initAttributeCapacity,this.initAttributeIncrement);
                index=0;
                reader.mark(1);
                while(reader.read()>0){
                    reader.reset();
                    String attributeName=reader.readLine();
                    if(attributeName!=null){
                        attributeName=attributeName.trim();
                        if(attributeName.length()>0){
                            this.attributes.add(index,new Attribute(attributeName));
                            this.log("Attribute"+index+"="+attributeName);
                            index++;
                        }
                    }
                    /* mark current position because read() call jumps ahead one character */
                    reader.mark(1);
                }
                /* no more elements will be added to the vector, trim it */
                this.attributes.trimToSize();
            } catch (Exception x){ 
                /* log error and quit */
                this.log(x.toString());
                this.updateStatus(this.errorBadNamesFile);
                this.error(x);
                this.endProgram();
            }
        }
        this.updateStatus("Done processing "+dataSetName+".");
    }
    
    /** reads and parses training file
     * @param dataSetName name of data set to use
     */    
    private void readTrainFile(String dataSetName){
        this.updateStatus("Processing "+dataSetName+"..");
        this.log("Reading train file"); 
        this.trainingData=new Vector(10,20);
        int index;
        int elementCount=this.attributes.size();
        this.log("elementCount="+elementCount);
        try{
            LineNumberReader reader=new LineNumberReader(new FileReader(dataSetName));
            /* read each line of file, mark current position because read() call jumps ahead one character */
            reader.mark(1);
            while(reader.read()>0){
                reader.reset();
                String currentLine=reader.readLine();
                if(currentLine!=null){
                    DataInstance dataInstance=new DataInstance(elementCount);
                    StringTokenizer tokenizer=new StringTokenizer(currentLine.trim(),this.fieldDelimeter);
                    index=0;
                    while(tokenizer.hasMoreTokens()){
                        String token=tokenizer.nextToken();
                        if(token.endsWith(".")){
                            token=token.substring(0,token.length()-1);
                        }
                        if(index==elementCount){
                            dataInstance.setClassification(token);
                        } else{
                            dataInstance.setData(index,Double.valueOf(token).doubleValue());
                            index++;
                        }
                    }
                    /* add dataInstance to vector */
                    this.trainingData.add(dataInstance);
                }
                /* mark current position because read() call jumps ahead one character */
                reader.mark(1);
            }
        } catch (Exception x){ 
            /* log error and quit */
            this.log(x.toString());
            this.updateStatus(this.errorFileNotFound+dataSetName);
            this.error(x);
            this.endProgram();
        }
        /* done reading, time to build decision tree */
        this.updateStatus(this.statusBuildTree);
        this.trainingData.trimToSize();
        this.initialTree=new DecisionTree(this.attributes,trainingData,this.classifications);
        this.log("Size of trainingData="+trainingData.size());
        this.updateStatus(this.statusTestingAccuracy);
        this.percentCorrectInitialTreeTrain=this.initialTree.testAccuracy(trainingData);
        this.updateStatus("Done processing "+dataSetName+".");
    }

    /** reads and parses prune file
     * @param dataSetName name of data set to use
     */    
    private void readPruneFile(String dataSetName){
        this.updateStatus("Processing "+dataSetName+"..");
        this.log("Reading prune file"); 
        Vector pruneData=new Vector(10,20);
        int index;
        int elementCount=this.attributes.size();
        this.log("elementCount="+elementCount);
        try{
            LineNumberReader reader=new LineNumberReader(new FileReader(dataSetName));
            /* read each line of file, mark current position because read() call jumps ahead one character */
            reader.mark(1);
            while(reader.read()>0){
                reader.reset();
                String currentLine=reader.readLine();
                if(currentLine!=null){
                    DataInstance dataInstance=new DataInstance(elementCount);
                    StringTokenizer tokenizer=new StringTokenizer(currentLine.trim(),this.fieldDelimeter);
                    index=0;
                    while(tokenizer.hasMoreTokens()){
                        String token=tokenizer.nextToken();
                        if(token.endsWith(".")){
                            token=token.substring(0,token.length()-1);
                        }
                        if(index==elementCount){
                            dataInstance.setClassification(token);
                        } else{
                            dataInstance.setData(index,Double.valueOf(token).doubleValue());
                            index++;
                        }
                    }
                    /* add dataInstance to vector */
                    pruneData.add(dataInstance);
                }
                /* mark current position because read() call jumps ahead one character */
                reader.mark(1);
            }
        } catch (Exception x){ 
            /* log error and quit */
            this.log(x.toString());
            this.updateStatus(this.errorFileNotFound+dataSetName);
            this.error(x);
            this.endProgram();
        }
        /* done reading, time to prune decision tree */
        this.updateStatus(this.statusPruning);
        /* copy current decision tree to separate pruning tree, so the two can be compared */
        this.prunedTree=(DecisionTree)this.initialTree.clone();
        pruneData.trimToSize();
        this.log("Size of pruneData="+pruneData.size());
        this.prunedTree.prune(pruneData);
        this.updateStatus(this.statusTestingAccuracy);
        this.percentCorrectInitialTreePrune=this.initialTree.testAccuracy(pruneData);
        this.percentCorrectPrunedTreePrune=this.prunedTree.testAccuracy(pruneData);
        pruneData.clear();
        this.percentCorrectPrunedTreeTrain=this.prunedTree.testAccuracy(this.trainingData);
        this.trainingData.clear();
        this.updateStatus("Done processing "+dataSetName+".");
    }
    
    /** reads and parses test file
     * @param dataSetName name of data set to use
     */    
    private void readTestFile(String dataSetName){
        this.updateStatus("Processing "+dataSetName+"& testing accuracy of trees, this may take a minute..");
        Vector testData=new Vector(10,20);
        int index;
        int elementCount=this.attributes.size();
        this.log("elementCount="+elementCount);
        try{
            LineNumberReader reader=new LineNumberReader(new FileReader(dataSetName));
            /* read each line of file, mark current position because read() call jumps ahead one character */
            reader.mark(1);
            while(reader.read()>0){
                reader.reset();
                String currentLine=reader.readLine();
                if(currentLine!=null){
                    DataInstance dataInstance=new DataInstance(elementCount);
                    StringTokenizer tokenizer=new StringTokenizer(currentLine.trim(),this.fieldDelimeter);
                    index=0;
                    while(tokenizer.hasMoreTokens()){
                        String token=tokenizer.nextToken();
                        if(token.endsWith(".")){
                            token=token.substring(0,token.length()-1);
                        }
                        if(index==elementCount){
                            dataInstance.setClassification(token);
                        } else{
                            dataInstance.setData(index,Double.valueOf(token).doubleValue());
                            index++;
                        }
                    }
                    /* add dataInstance to vector */
                    testData.add(dataInstance);
                }
                /* mark current position because read() call jumps ahead one character */
                reader.mark(1);
            }
        } catch (Exception x){ 
            /* log error and quit */
            this.log(x.toString());
            this.updateStatus(this.errorFileNotFound+dataSetName);
            this.error(x);
            this.endProgram();
        }
        this.updateStatus(this.statusTestingAccuracy);
        testData.trimToSize();
        this.percentCorrectInitialTreeTest=this.initialTree.testAccuracy(testData);
        testData.clear();
        this.updateStatus("Done processing "+dataSetName+".");
    }
    
    /** returns the percect of correct classifications made on the initial tree for the test data set
     * @return percect of correct classifications made on the initial tree for the test data set
     */
    public double getPercentCorrectInitialTreeTest(){
        return(this.percentCorrectInitialTreeTest);
    }
    
    /** returns the percect of correct classifications made on the pruned tree for the test data set
     * @return percect of correct classifications made on the pruned tree for the test data set
     */
    public double getPercentCorrectPrunedTreeTest(){
        return(this.percentCorrectPrunedTreeTest);
    }
    
    /** returns the percect of correct classifications made on the initial tree for the pruning data set
     * @return percect of correct classifications made on the initial tree for the pruning data set
     */
    public double getPercentCorrectInitialTreePrune(){
        return(this.percentCorrectInitialTreePrune);
    }
    
    /** returns the percect of correct classifications made on the pruned tree for the pruning data set
     * @return percect of correct classifications made on the pruned tree for the pruning data set
     */
    public double getPercentCorrectPrunedTreePrune(){
        return(this.percentCorrectPrunedTreePrune);
    }
    
    /** returns the percect of correct classifications made on the initial tree for the training data set
     * @return percect of correct classifications made on the initial tree for the training data set
     */
    public double getPercentCorrectInitialTreeTrain(){
        return(this.percentCorrectInitialTreeTrain);
    }
    
    /** returns the percect of correct classifications made on the pruned tree for the training data set
     * @return percect of correct classifications made on the pruned tree for the training data set
     */
    public double getPercentCorrectPrunedTreeTrain(){    
        return(this.percentCorrectPrunedTreeTrain);
    }
    
    /** override this 
     *  @param message the message to display
     */
    public void updateStatus(String message){
        /* override this */
    }
    
    /** override this
     */
    public void done(){
        /* override this */
    }

    /** override this 
     * @param x the error that occured
     */
    public void error(Exception x){
        /* override this */
    }
    
    /** add a line to the execution log file
     * @param line the line to append
     */
    private void log(String line){
        if(this.createLog){
            if(this.logFile!=null){
                this.logFile.log(line);
            }
        }
    }
    
    /** serialize the initial tree
     * @param fileName the file to save the tree to
     */
    public void saveInitialTree(String fileName){
        try{
            ObjectOutputStream oOut=new ObjectOutputStream(new FileOutputStream(fileName));
            oOut.writeObject(this.initialTree);
            oOut.flush();
            oOut.close();
        } catch(IOException iox){
            iox.printStackTrace();
        }
    }
    
    /** serialize the pruned tree
     * @param fileName the file to save the tree to
     */
    public void savePrunedTree(String fileName){
        try{
            ObjectOutputStream oOut=new ObjectOutputStream(new FileOutputStream(fileName));
            oOut.writeObject(this.prunedTree);
            oOut.flush();
            oOut.close();
        } catch(IOException iox){
            iox.printStackTrace();
        }    
    }

    /** deserialize the initial tree
     * @param fileName the file to read the tree from
     */
    public void loadInitialTree(String fileName){
        try{
            ObjectInputStream oIn=new ObjectInputStream(new FileInputStream(fileName));
            this.initialTree=(DecisionTree)oIn.readObject();
            oIn.close();
        } catch(IOException iox){
            iox.printStackTrace();
        } catch(ClassNotFoundException cnfx){
            cnfx.printStackTrace();
        }
    }
    
    /** deserialize the pruned tree
     * @param fileName the file to read the tree from
     */
    public void loadPrunedTree(String fileName){
        try{
            ObjectInputStream oIn=new ObjectInputStream(new FileInputStream(fileName));
            this.prunedTree=(DecisionTree)oIn.readObject();
            oIn.close();
        } catch(IOException iox){
            iox.printStackTrace();
        } catch(ClassNotFoundException cnfx){
            cnfx.printStackTrace();
        }    
    }    
    
    /** returns a string representation of the object
     * @return string representation of the object
     */
    public String toString(){
        String tostring=new String(this.getClass().getName());
        return(tostring);
    }    
}