package com.huguesjohnson.jail.neuralNetwork;

/*
neuralNetwork - Package of neural network objects
BackPropagationLayer.java - Layer for a backpropation network
Copyright  (C) 2000-2003 Hugues Johnson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

import com.huguesjohnson.jail.neuralNetwork.Layer;
import com.huguesjohnson.jail.neuralNetwork.Perceptron;

/** class BackPropagationLayer
 */
public class BackPropagationLayer extends Layer{

    /* contructor(s) */
    
    /** default constructor
     */
    public BackPropagationLayer(){
        super();
    }
    
    /** constructor
     * @param layerSize the number of neurons in the layer
     * @param inputSize the number of inputs to the layer
     * @param punishRate the punish rate for every neuron in the layer
     * @param rewardRate the reward rate for every neuron in the layer
     * @param errorThreshold the error threshold for every neuron in the layer
     * @param minWeight minimum weight on a neuron link
     * @param maxWeight maximum weight on a neuron link
     */
    public BackPropagationLayer(int layerSize,int inputSize,double punishRate,double rewardRate,double errorThreshold,double minWeight,double maxWeight){
        super();
        for(int index=0;index<layerSize;index++){
            Perceptron neuron=new Perceptron(punishRate,rewardRate,errorThreshold);
            neuron.randomizeWeights(inputSize,minWeight,maxWeight);
            this.insertNeuron(neuron);
        }
    }
    
    /* overrides of Layer methods */
    
    /** method invoked when run is complete
     */
    public void runComplete(){
        /* override this in implementing class */
    }
    
    /** move actual output closer to desired output 
     * by default find neuron with greatest error and adjust it
     */
    public void train(){
        int size=this.getSize();
        double highError=0.0D;
        int highErrorIndex=-1;
        try{
            for(int index=0;index<size;index++){
                Neuron neuron=(this.getNeuron(index));
                if(neuron.getDoneTraining()==false){
                    double currentError=Math.abs(neuron.getDesiredOutput()-this.getOutput(index));
                    if(currentError>highError){
                        highError=currentError;
                        highErrorIndex=index;
                    }
                }
            }
            if(highErrorIndex>=0){
                Neuron neuron=(Neuron)this.getNeuron(highErrorIndex);
                neuron.adjustWeights();
                /* is this neuron done training? */
                double newError=Math.abs(neuron.getDesiredOutput()-this.getOutput(highErrorIndex));
                if(newError<neuron.getErrorThreshold()){
                    neuron.setDoneTraining(true);
                }
                this.setNeuron(highErrorIndex,neuron);
            }
        } catch(LayerException lx){
            lx.printStackTrace();
        }
    }
    
    /** computes the output for the layer
     */
    public void computeOutput(){
        int size=this.getSize();
        this.initOutput();
        try{
            for(int index=0;index<size;index++){
                this.setOutput(index,(this.getNeuron(index).activation()));
            }
        } catch(LayerException lx){
            lx.printStackTrace();
        }
    }    
    
    /* methods unique to BackPropagationLayer */
    
}