package com.huguesjohnson.jail.neuralNetwork;

/*
neuralNetwork - Package of neural network objects
KohonenNeuron.java - Extention of neuron
Copyright  (C) 2000-2003 Hugues Johnson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

import com.huguesjohnson.jail.neuralNetwork.Neuron;
import com.huguesjohnson.jail.neuralNetwork.NeuronException;

/** KohonenNeuron
 */
public class KohonenNeuron extends Neuron{

    /* constructors */

    /** default constructor
     */
    public KohonenNeuron(){
        super();
    }

    /** constructor
     * @param desiredOutput the desired output of the neuron
     * @param punishRate the punish rate of the neuron
     * @param rewardRate the reward rate of the neuron
     * @param errorThreshold the error threshold of the neuron
     */    
    public KohonenNeuron(double desiredOutput,double punishRate,double rewardRate,double errorThreshold){
        super(desiredOutput,punishRate,rewardRate,errorThreshold);
    }

    /** constructor
     * @param input the input to the neuron
     * @param weights the weights on the links to the neuron
     * @param desiredOutput the desired output of the neuron
     * @param punishRate the punish rate of the neuron
     * @param rewardRate the reward rate of the neuron
     * @param errorThreshold the error threshold of the neuron
     */       
    public KohonenNeuron(double[] input,double weights[],double desiredOutput,double punishRate,double rewardRate,double errorThreshold){
        super(input,weights,desiredOutput,punishRate,rewardRate,errorThreshold);
    }

    /** constructor
     * @param punishRate the punish rate of the neuron
     * @param rewardRate the reward rate of the neuron
     * @param errorThreshold the error threshold of the neuron
     */    
    public KohonenNeuron(double punishRate,double rewardRate,double errorThreshold){
        super(punishRate,rewardRate,errorThreshold);
    }    
    
    /* overrides of neuron methods */
    
    /** activation function
     * @return activation for the neuron
     */
    public double activation(){
        return(this.getZ());
    }    
    
    /** adjust weights for the neuron
     */
    public void adjustWeights(){
        double actualOutput=this.activation();
        double desiredOutput=this.getDesiredOutput();
        double errorThreshold=this.getErrorThreshold();
        double actualError=Math.abs(desiredOutput-actualOutput);
        /* if actualError is less than errorThreshold there is no need to adjust weights */
        if(actualError>errorThreshold){
            int length=this.getWeightCount();
            double bestError=actualError;
            double bestValue=0.0D;
            int bestIndex=0;
            int weightIndex=0;
            try{
                /* if the error is reduced below the threshold stop searching */
                while((weightIndex<length)&&(bestError>errorThreshold)){
                    double originalWeight=this.getWeight(weightIndex);
                    double input=this.getInput(weightIndex);
                    double linkOutput=(originalWeight*input);
                    /* if weight is below ? flip sign on weight */
                    if(Math.abs(originalWeight)<0.0001D){
                        this.setWeight(weightIndex,(originalWeight*(-1.0001D)));
                    } else{
                        /* should this weight be rewarded or punished? */
                        if(actualOutput>desiredOutput){ /* need to decrease actual */
                            if(linkOutput<0.0D){
                                this.rewardWeight(weightIndex);
                            } else{/* (link>0) */
                                this.punishWeight(weightIndex);
                            }
                        } else{ /* desired>actual, need to increase actual */
                            if(linkOutput<0.0D){
                                this.punishWeight(weightIndex);
                            } else{ /* link>0 */
                                this.rewardWeight(weightIndex);
                            }
                        }
                    }
                    /* test if improvement was made */
                    actualOutput=this.activation();
                    actualError=Math.abs(desiredOutput-actualOutput);
                    if(actualError<bestError){
                        bestError=actualError;
                        bestIndex=weightIndex;
                        bestValue=this.getWeight(weightIndex);
                    }
                    /* restore the previous value of the weight */
                    this.setWeight(weightIndex,originalWeight);
                    /* increment weightIndex */
                    weightIndex++;
                }
                /* set the new weight value */
                this.setWeight(bestIndex,bestValue);
            } catch(NeuronException nx){
                nx.printStackTrace();
            }
        }
    }        
    
    /* methods specific to KohonenNeuron */
}