package com.huguesjohnson.jail.neuralNetwork;

/*
neuralNetwork - Package of neural network objects
Perceptron.java - Extention of neuron
Copyright  (C) 2000-2003 Hugues Johnson

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software 
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

import java.io.Serializable;
import java.lang.Math;
import java.lang.StringBuffer;
import com.huguesjohnson.jail.neuralNetwork.Neuron;
import com.huguesjohnson.jail.neuralNetwork.NeuronException;

/** Perceptron
 */
public class Perceptron extends Neuron implements Serializable{
    
    /* constructors */

    /** default constructor
     */
    public Perceptron(){
        super();
    }
    
    /** constructor
     * @param desiredOutput the desired output of the neuron
     * @param punishRate the punish rate of the neuron
     * @param rewardRate the reward rate of the neuron
     * @param errorThreshold the error threshold of the neuron
     */        
    public Perceptron(double desiredOutput,double punishRate,double rewardRate,double errorThreshold){
        super(desiredOutput,punishRate,rewardRate,errorThreshold);
    }
    
    /** constructor
     * @param input the input to the neuron
     * @param weights the weights on the links to the neuron
     * @param desiredOutput the desired output of the neuron
     * @param punishRate the punish rate of the neuron
     * @param rewardRate the reward rate of the neuron
     * @param errorThreshold the error threshold of the neuron
     */         
    public Perceptron(double[] input,double weights[],double desiredOutput,double punishRate,double rewardRate,double errorThreshold){
        super(input,weights,desiredOutput,punishRate,rewardRate,errorThreshold);
    }

    /** constructor
     * @param punishRate the punish rate of the neuron
     * @param rewardRate the reward rate of the neuron
     * @param errorThreshold the error threshold of the neuron
     */        
    public Perceptron(double punishRate,double rewardRate,double errorThreshold){
        super(punishRate,rewardRate,errorThreshold);
    }    
    
    /* overrides of Neuron methods */

    /** activation function for neuron
     * @return activation value
     */
    public double activation(){
        return(this.sigmoid(this.getZ()));
    }    
    
    /** adjust weights for the neuron
     */
    public void adjustWeights(){
        double actualOutput=this.activation();
        double desiredOutput=this.getDesiredOutput();
        double errorThreshold=this.getErrorThreshold();
        double actualError=Math.abs(desiredOutput-actualOutput);
        double startError=actualError;
        /* if actualError is less than errorThreshold there is no need to adjust weights */
        if(actualError>errorThreshold){
            int length=this.getWeightCount();
            double bestError=actualError;
            double bestValue=0.0D;
            int bestIndex=0;
            int weightIndex=0;
            try{
                /* if the error is reduced below the threshold stop searching */
                while((weightIndex<length)&&(bestError>errorThreshold)){
                    double originalWeight=this.getWeight(weightIndex);
                    double input=this.getInput(weightIndex);
                    double linkOutput=(originalWeight*input);
                    /* if weight is below ? try reversing sign on weight instead of punish or reward */
                    if(Math.abs(originalWeight)<0.0001D){
                        this.setWeight(weightIndex,(originalWeight*(-1.0001D)));
                    } else{
                        /* should this weight be rewarded or punished? */
                        if(actualOutput>desiredOutput){ /* need to decrease actual */
                            if(linkOutput<0.0D){
                                this.rewardWeight(weightIndex);
                            } else{/* (link>0) */
                                this.punishWeight(weightIndex);
                            }
                        } else{ /* desired>actual, need to increase actual */
                            if(linkOutput<0.0D){
                                this.punishWeight(weightIndex);
                            } else{ /* link>0 */
                                this.rewardWeight(weightIndex);
                            }
                        }
                    }
                    /* test if improvement was made */
                    actualOutput=this.activation();
                    actualError=Math.abs(desiredOutput-actualOutput);
                    if(actualError<bestError){
                        bestError=actualError;
                        bestIndex=weightIndex;
                        bestValue=this.getWeight(weightIndex);
                    }
                    /* restore the previous value of the weight */
                    this.setWeight(weightIndex,originalWeight);
                    /* increment weightIndex */
                    weightIndex++;
                }
                /* if an improvement was made - set the new weight value */
                if(bestError<startError){
                    this.setWeight(bestIndex,bestValue);
                } else{ 
                    /* no improvement made */
                }
            } catch(NeuronException nx){
                nx.printStackTrace();
            }
        }
    }        
    
    /* methods specific to Perceptron class */
    
    /** sigmoid function
     * @param value input to the sigmoid function
     * @return double between -1.0 and 1.0
     */
    private double sigmoid(double value){
        double f=(2.0D/(1+Math.exp(-1.0D*value)))-1.0D;
        return(f);
    }    
}